﻿source("dataS/scripts/vehicles/specializations/SetTurnedOnEvent.lua");

BTSC = {};

function BTSC.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Fillable, specializations);
end;

function BTSC:load(xmlFile)

    self.mixerWagonSuperAllowFillType = self.allowFillType;
    self.allowFillType = Utils.overwrittenFunction(self.allowFillType, BTSC.allowFillType);
    self.getFillLevel = Utils.overwrittenFunction(self.getFillLevel, BTSC.getFillLevel);
    self.setFillLevel = Utils.overwrittenFunction(self.setFillLevel, BTSC.setFillLevel);
    self.resetFillLevelIfNeeded = Utils.overwrittenFunction(self.resetFillLevelIfNeeded, BTSC.resetFillLevelIfNeeded);
    self.BTSCBaleTriggerCallback = BTSC.BTSCBaleTriggerCallback;

    self.BTSCFillTypes = {};
    self.fillTypeToBTSCFillType = {};
    local i=0;
    while true do
        local baseName = string.format("vehicle.BTSCFillTypes.BTSCFillType(%d)", i);
        if not hasXMLProperty(xmlFile, baseName) then
            break;
        end;
        local entry = {};
        entry.fillTypes = {};

        local j=0;
        while true do
            local fillTypeKey = baseName..string.format(".fillType(%d)", j);
            if not hasXMLProperty(xmlFile, fillTypeKey) then
                break;
            end;
            local fillType = getXMLString(xmlFile, fillTypeKey.."#fillType");
            if fillType ~= nil then
                local fillTypeInt = Fillable.fillTypeNameToInt[fillType];
                if fillTypeInt ~= nil then
                    if self.fillTypeToBTSCFillType[fillTypeInt] == nil then
                        entry.fillTypes[fillTypeInt] = true;
                        self.fillTypeToBTSCFillType[fillTypeInt] = entry;
                    else
                        print("Warning: Invalid BTSCFillType '"..fillType.."' used multiple times in '"..self.configFileName.."'.");
                    end
                else
                    print("Warning: Invalid BTSCFillType '"..fillType.."' in '"..self.configFileName.. "'.");
                end
            end

            j = j+1;
        end
        entry.minPercentage = Utils.getNoNil(getXMLFloat(xmlFile, baseName.."#minPercentage"), 0)*0.01;
        entry.maxPercentage = Utils.getNoNil(getXMLFloat(xmlFile, baseName.."#maxPercentage"), 100)*0.01;
        entry.fillLevel = 0;

        local name = Utils.getNoNil(getXMLString(xmlFile, baseName.."#name"), "unknown");
        entry.name = name;
        local i18n = g_i18n;
        if self.customEnvironment ~= nil then
            i18n = _G[self.customEnvironment].g_i18n;
        end;
        entry.nameI18N = i18n:getText(name);
        if next(entry.fillTypes) ~= nil then
            entry.fillGreenOverlay = Overlay:new("", "dataS2/menu/hud/vehicleHUD_barGreen.png", 0, 0, 1, 1);
            entry.fillRedOverlay = Overlay:new("", "dataS2/menu/hud/vehicleHUD_barRed.png", 0, 0, 1, 1);

            table.insert(self.BTSCFillTypes, entry);
        end
        i = i+1;
    end;


    if self.isServer then
        self.BTSCBaleTriggerId = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.BTSCBaleTrigger#index"));
        if self.BTSCBaleTriggerId ~= nil then
            addTrigger(self.BTSCBaleTriggerId, "BTSCBaleTriggerCallback", self);
        end
    end

    self.mixingActiveTimerMax = 5000;
    self.mixingActiveTimer = 0;

    local fillTypeOverlayWidth = 0.18;
    local fillTypeOverlayHeight = 0.028;
    self.fillTypeOverlayLineSpacing = 0.034;
    self.BTSCFillTypeFrameOverlay = Overlay:new("", "dataS2/menu/hud/mixerWagonHUD_frame.png", 0, 0, fillTypeOverlayWidth, fillTypeOverlayHeight);
    self.BTSCFillTypeRangeMarkerLeftOverlay = Overlay:new("", "dataS2/menu/hud/mixerWagonHUD_rangeMarkerLeft.png", 0, 0, fillTypeOverlayWidth*20/512.0, fillTypeOverlayHeight);
    self.BTSCFillTypeRangeMarkerRightOverlay = Overlay:new("", "dataS2/menu/hud/mixerWagonHUD_rangeMarkerLeft.png", 0, 0, fillTypeOverlayWidth*20/512.0, fillTypeOverlayHeight);
    self.BTSCFillTypeRangeMarkerRightOverlay:setInvertX(true);

    self.mixerWagonLastPickupTime = -10000;

    -- disable fill level sync and save of Fillable
    self.synchronizeFillLevel = false;


    self.mixerWagonDirtyFlag = self:getNextDirtyFlag();
end;

function BTSC:delete()
    if self.isServer then
        if self.BTSCBaleTriggerId ~= nil then
            removeTrigger(self.BTSCBaleTriggerId);
        end
    end
    for _,entry in pairs(self.BTSCFillTypes) do
        entry.fillGreenOverlay:delete();
        entry.fillRedOverlay:delete();
    end
end;

function BTSC:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
    self.fillLevel = 0;
    for i=1, table.getn(self.BTSCFillTypes) do
        local fillTypeKey = key..string.format(".BTSCFillType%d#fillLevel", i);
        local fillLevel = getXMLFloat(xmlFile, fillTypeKey);
        if fillLevel == nil then
            fillLevel = 0;
        end
        self:setFillLevel(fillLevel, next(self.BTSCFillTypes[i].fillTypes));
    end
    return BaseMission.VEHICLE_LOAD_OK;
end;

function BTSC:getSaveAttributesAndNodes(nodeIdent)
    local nodes = "";
    for i=1, table.getn(self.BTSCFillTypes) do
        if i>1 then
            nodes = nodes.."\n";
        end;
        local fillLevel = self.BTSCFillTypes[i].fillLevel;
        nodes = nodes.. nodeIdent..string.format('<BTSCFillType%d', i)..' fillLevel="'..fillLevel..'" />';
    end;
    return "", nodes;
end;

function BTSC:readStream(streamId, connection)
    self.fillLevel = 0;
    for i=1,table.getn(self.BTSCFillTypes) do
        local fillLevel = streamReadFloat32(streamId);
        self:setFillLevel(fillLevel, next(self.BTSCFillTypes[i].fillTypes));
    end
end

function BTSC:writeStream(streamId, connection)
    for i=1,table.getn(self.BTSCFillTypes) do
        streamWriteFloat32(streamId, self.BTSCFillTypes[i].fillLevel);
    end
end

function BTSC:readUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then
            self.mixingActiveTimer = streamReadUIntN(streamId, 4)/15 * self.mixingActiveTimerMax;
            for i=1,table.getn(self.BTSCFillTypes) do
                local fillLevel = streamReadUInt8(streamId)/255*self.capacity;
                self:setFillLevel(fillLevel, next(self.BTSCFillTypes[i].fillTypes));
            end
        end;
    end
end

function BTSC:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.mixerWagonDirtyFlag) ~= 0) then
            streamWriteUIntN(streamId, math.floor(self.mixingActiveTimer/self.mixingActiveTimerMax*15), 4);
            for i=1,table.getn(self.BTSCFillTypes) do
                local percent = 0;
                if self.capacity ~= 0 then
                    percent = Utils.clamp(self.BTSCFillTypes[i].fillLevel / self.capacity, 0, 1);
                end;
                streamWriteUInt8(streamId, math.floor(percent*255));
            end
        end
    end
end;

function BTSC:mouseEvent(posX, posY, isDown, isUp, button)
end;

function BTSC:keyEvent(unicode, sym, modifier, isDown)
end;

function BTSC:update(dt)
    if self.isActive then
        if self.mixerWagonNeedsTurnOn then
            if self:getIsActiveForInput() then
                if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA) then
                    self:setIsTurnedOn(not self.isTurnedOn);
                end
            end
        end
    end
end;

function BTSC:updateTick(dt)
end;

function BTSC:draw()
end;

-- overwrite Fillable.resetFillLevelIfNeeded
function BTSC:resetFillLevelIfNeeded(superFunc, fillType)
    local BTSCFillType = self.fillTypeToBTSCFillType[fillType];
    if BTSCFillType ~= nil then
        -- nothing to do
        return;
    end
    superFunc(self, fillType);
end;

function BTSC:allowFillType(superFunc, fillType, allowEmptying)
    local BTSCFillType = self.fillTypeToBTSCFillType[fillType];
    if BTSCFillType ~= nil then
        return true;
    end

    return superFunc(self, fillType, allowEmptying);
end

function BTSC:getFillLevel(superFunc, fillType)
    local BTSCFillType = self.fillTypeToBTSCFillType[fillType];
    if BTSCFillType ~= nil then
        return BTSCFillType.fillLevel;
    end
    return superFunc(self, fillType);
end

function BTSC:setFillLevel(superFunc, fillLevel, fillType, force)

    local BTSCFillType = self.fillTypeToBTSCFillType[fillType];
    if BTSCFillType == nil then

        if self:allowFillType(fillType) and fillLevel < self.fillLevel and self.fillLevel > 0 then
            -- remove values from all fill types such that the ratio doesn't change
            fillLevel = Utils.clamp(fillLevel, 0, self.capacity);
            local delta = fillLevel - self.fillLevel;
            local newFillLevel = 0;
            for _, entry in pairs(self.BTSCFillTypes) do
                entry.fillLevel = math.max(entry.fillLevel + delta * entry.fillLevel / self.fillLevel, 0);
                newFillLevel = newFillLevel + entry.fillLevel;
            end

            self:raiseDirtyFlags(self.mixerWagonDirtyFlag);

            superFunc(self, newFillLevel, self.currentFillType, force);
        end

        return;

    end

    local isSingleFilled = (self.fillLevel == BTSCFillType.fillLevel) and self:mixerWagonSuperAllowFillType(fillType);

    local limitedFillLevel = fillLevel;
    limitedFillLevel = math.min(limitedFillLevel, self.capacity - self.fillLevel + BTSCFillType.fillLevel);
    -- dont check for validity any more. player can always empty a bit of the forage mixing at the feeding trough and fill up later
    --[[if not isSingleFilled then
        -- make sure that we are not above the max percentage
        limitedFillLevel = math.min(limitedFillLevel, BTSCFillType.maxPercentage * self.capacity);
        -- check that all other minPercentages can still be fullfilled
        local othersMinPercentagesFillLevel = 0;
        for _, entry in pairs(self.BTSCFillTypes) do
            if entry ~= BTSCFillType then
                othersMinPercentagesFillLevel = othersMinPercentagesFillLevel + entry.minPercentage * self.capacity
            end
        end
        limitedFillLevel = math.min(limitedFillLevel, self.capacity-othersMinPercentagesFillLevel);
    end]]
    limitedFillLevel = Utils.clamp(limitedFillLevel, 0, self.capacity);

    if limitedFillLevel ~= BTSCFillType.fillLevel then
        BTSCFillType.fillLevel = limitedFillLevel;
        self:raiseDirtyFlags(self.mixerWagonDirtyFlag);
    end
	
    -- calculate new fill level and forage is ok
    --local newFillLevel = 0;
    --local isForageOk = true;
    --for _, entry in pairs(self.BTSCFillTypes) do
    --    newFillLevel = newFillLevel + entry.fillLevel;
    --end
    --for _, entry in pairs(self.BTSCFillTypes) do
    --    if entry.fillLevel < entry.minPercentage * newFillLevel-0.01 or entry.fillLevel > entry.maxPercentage * newFillLevel + 0.01 then
    --        isForageOk = false;
    --        break;
    --    end
    --end

    --local newFillType = Fillable.FILLTYPE_FORAGE_MIXING;
    --if isForageOk then
    --    newFillType = Fillable.FILLTYPE_WHEAT_WINDROW;
    --else
    --    if isSingleFilled then
    --        newFillType = fillType;
    --    else
            -- maybe we are single filled now
    --        local singleFilledFillType;
    --        for _, entry in pairs(self.BTSCFillTypes) do
    --            if entry.fillLevel == newFillLevel then
    --                isSingleFilled = true;
    --                newFillType = next(entry.fillTypes);
    --                break;
    --            end
    --        end
    --    end
    --end

    superFunc(self, fillLevel, fillType, true);
end

function BTSC:BTSCBaleTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter then
        -- this happens if a compound child of a deleted compound is entering
        if otherActorId ~= 0 then
            local object = g_currentMission:getNodeObject(otherActorId);
            if object ~= nil then
                if object:isa(Bale) then
                    local fillLevel = object:getFillLevel();
                    local fillType = object:getFillType();
                    if self:allowFillType(fillType, false) then
                        self:setFillLevel(self:getFillLevel(fillType)+fillLevel, fillType);
                    end
                    object:delete();

                    self.mixingActiveTimer = self.mixingActiveTimerMax;
                    self:raiseDirtyFlags(self.mixerWagonDirtyFlag);
                end;
            end;
        end;
    end;
end;